/*****************************************************************************
 *          Copyright: (C) Copyright David Goodwin, 2014
 *            License: GNU General Public License
 *****************************************************************************
 *
 *   This is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This software is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this software; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ****************************************************************************/

#include "databasedatasource.h"
#include "settings.h"

#include "dbutil.h"

#include <QSqlDatabase>
#include <QStringList>
#include <QSqlQuery>
#include <QSqlError>
#include <QtDebug>
#include <QSqlRecord>

#include <cmath>
#include <algorithm>

DatabaseDataSource::DatabaseDataSource(QObject *parent) :
    AbstractDataSource(parent)
{
    collecting = false;
    isZxWeatherV2Schema = false;
}

DatabaseDataSource::~DatabaseDataSource() {
    closeDevice();
}

bool DatabaseDataSource::openDevice() {

    Settings& settings = Settings::getInstance();

    if (QSqlDatabase::isDriverAvailable("QPSQL")) {
        QSqlDatabase db = QSqlDatabase::addDatabase("QPSQL");
        db.setHostName(settings.databaseHostName());
        db.setPort(settings.databasePort());
        db.setDatabaseName(settings.databaseName());
        db.setUserName(settings.databaseUsername());
        db.setPassword(settings.databasePassword());
        zxwStationCode = settings.stationCode();
        useIndoorSensors = settings.useIndoorSensors();

        bool ok = db.open();
        if (!ok) {
            emit error(tr("Database Connect Failed"),
                       tr("Failed to connect to the database. Check database "
                       "settings."));
            return false;
        }

        auto compatibility = DbUtil::checkDatabaseCompatibility(db, true);
        if (compatibility == DbUtil::DC_Compatible) {
            auto type = DbUtil::getDatabaseType(db);
            isZxWeatherV2Schema = (type == DbUtil::DT_ZXW_V2);
            isZxWeather = (type == DbUtil::DT_ZXW_V1 || type == DbUtil::DT_ZXW_V2);
        }
        else if (compatibility == DbUtil::DC_Incompatible_Version) {
            QString ver = DbUtil::getMinimumAppVersion(db);
            emit error(tr("Incompatbile Database"),
                       QString(tr("The configured data source database is not compatbile "
                       "with this version of TempView. Please upgrade to TempView version"
                       " %1 or newer.")).arg(ver));
            return false;
        }
        else {
            emit error(tr("Incompatible Database"),
                       tr("The configured database is incompatible with this application."
                       " Check database configuration."));
            return false; // Incompatible database.
        }

        if (isZxWeatherV2Schema) {
            int id = getStationId();

            if (id == -1)
                return false; // Station code not configured.
        }

        // Database looks ok so far. Try subscribing to notifications.
        auto driver = db.driver();
        connect(driver,
                SIGNAL(notification(QString,QSqlDriver::NotificationSource,QVariant)),
                this,
                SLOT(databaseNotification(QString,QSqlDriver::NotificationSource,QVariant)));

        // If the subscription is successfull we're done!
        if (!db.driver()->subscribeToNotification("live_data_updated")) {
            emit error(tr("Subscription Error"),
                       tr("Failed to subscribe to database. Check database configuration."));
            return false;
        }

        if (!db.driver()->subscribeToNotification("live_data_updated_v2")) {
            emit error(tr("Subscription Error"),
                       tr("Failed to subscribe to database. Check database configuration."));
            return false;
        }
        return true;

    } else {
        emit error(tr("Database Driver Not Found"),
                   tr("The Qt PostgreSQL database driver was not found."
                   " Unable to use remote PostgreSQL database as a"
                   " data source."));
        return false;
    }
}

bool DatabaseDataSource::closeDevice() {
    stopCollecting();
    auto db = QSqlDatabase::database();
    db.close();
    QSqlDatabase::removeDatabase(QSqlDatabase::defaultConnection);
    return true;
}

void DatabaseDataSource::startCollecting() {
    collecting = true;
}

void DatabaseDataSource::stopCollecting() {
    collecting = false;
}

void DatabaseDataSource::databaseNotification(const QString& name,
                          QSqlDriver::NotificationSource source,
                          const QVariant& payload) {
    qDebug() << "DB Notification Received" << name << source << payload;
    if (name == "live_data_updated" && (isZxWeatherV2Schema || isZxWeather)) {
        if (isZxWeatherV2Schema) {
            // Its a zxweather database. Payload is the code for the
            // station that just updated its live data. Check if its
            // the station we're interested in and pull the data out
            // of the database if it is.

            if (payload.toString() != zxwStationCode)
                return; // This is not the station we are looking for.

            QSqlQuery qry;
            if (useIndoorSensors) {
                qry.prepare("select indoor_temperature, indoor_relative_humidity from live_data where station_id = :stationId");
            } else {
                qry.prepare("select temperature, relative_humidity from live_data where station_id = :stationId");
            }
            qry.bindValue(":stationId", getStationId());
            qry.exec();
            if (qry.isActive() && qry.size() == 1) {
                qry.first();
                emit newTempHumData(qry.value(0).toFloat(), qry.value(1).toFloat());
                return;
            }
        } else if (isZxWeather){
            // zxweather v0.x database (schema v1). This database
            // only supports one station so just pull the data out
            // of the live_data table.

            QString queryText = "select temperature, relative_humidity from live_data";
            if (useIndoorSensors) {
                queryText = "select indoor_temperature, indoor_relative_humidity from live_data";
            }

            QSqlQuery qry(queryText);
            if (qry.isActive() && qry.size() == 1) {
                qry.first();
                emit newTempHumData(qry.value(0).toFloat(), qry.value(1).toFloat());
                return;
            }
        }
    } else if (name == "live_data_updated_v2") {
        // Broadcast from TempView v2.0+

        // TempView from v2.0 broadcasts live data as a number of comma
        // separated values. There will always be at least two values:
        // temperature and humidity. Humidity could be a '-' character to
        // indicate those sensors aren't preset. Temperature will always
        // contain a valid float.

        QStringList bits = payload.toString().split(",");

        float temperature, humidity;

        temperature = bits.at(0).toFloat();

        if (bits.at(1) == "-") {
            // No humidity data (therefore, no dewPoint data)
            emit newData(temperature);
        } else {
            humidity = bits.at(1).toFloat();
            emit newTempHumData(temperature, humidity);
        }

        return;
    }
}

int DatabaseDataSource::getStationId() {
    static int stationId = -1;
    static QString stationCode = "";

    QString code = Settings::getInstance().stationCode();

    if (code == stationCode) return stationId;

    stationId = -1;
    stationCode = "";

    QSqlQuery query;
    query.prepare("select station_id from station where code = :code");
    query.bindValue(":code", code);
    query.exec();

    if (!query.isActive()) {
        emit error(tr("Database Error"), query.lastError().databaseText());
        return -1;
    }
    else if (query.size() == 1) {
        query.first();
        stationId = query.value(0).toInt();
        stationCode = code;
    } else {
        emit error(tr("No Station Code Specified"),
                   tr("When using a zxweather database you must specify the"
                   " station code to subscribe to in the TempView configuration"
                   " file."));
        return -1;
    }

    qDebug() << "Station code" << code << "is ID" << stationId;

    return stationId;
}

float dewPoint(float temperature, float humidity) {
    float a = 17.271;
    float b = 237.7; // in °C
    float gamma, result;

    gamma = ((a * temperature) / (b + temperature)) + log(humidity / 100.0);
    result = (b * gamma) / (a - gamma);
    return result;
}

SampleSet DatabaseDataSource::fetchSamples(SampleColumns columns, QDateTime start, QDateTime end) {
    SampleSet s;

    QString tempColumn = "temperature";
    QString humidityColumn = "humidity";

    if (isZxWeather) {
        tempColumn = useIndoorSensors ? "indoor_temperature" : tempColumn;
        humidityColumn = useIndoorSensors ? "indoor_relative_humidity": "relative_humidity";
    }

    QString columnList = "select time_stamp ";

    if (columns.testFlag(SC_Temperature)) columnList += ", " + tempColumn + " ";
    if (columns.testFlag(SC_Humidity))    columnList += ", " + humidityColumn + " ";
    if (columns.testFlag(SC_DewPoint)) {
        if (isZxWeather && useIndoorSensors) {
            // zxweather doesn't store dew point for indoors - we'll have to
            // calculate it outselves. That means we'll need the temperature
            // and humidity columns in the query results. So add them if
            // they're not already in the column list.
            if (!columns.testFlag(SC_Temperature))
                columnList += ", " + tempColumn + " ";
            if (!columns.testFlag(SC_Humidity))
                columnList += ", " + humidityColumn + " ";

        } else {
            columnList += ", dew_point ";
        }
    }

    QString from = " from sample "
                   "where time_stamp >= :startTs "
                   "  and time_stamp <= :endTs ";

    QString queryText = columnList + from;

    if (isZxWeatherV2Schema)
        queryText += "  and station_id = :stationId ";

    queryText += "order by time_stamp";

    QSqlQuery query;
    query.prepare(queryText);
    query.bindValue(":startTs", start);
    query.bindValue(":endTs", end);

    if (isZxWeatherV2Schema)
        query.bindValue(":stationId", getStationId());

    bool result = query.exec();

    if (!result || !query.isActive()) {
        emit error(tr("Database Error"), query.lastError().driverText());
        return s;
    }

    while (query.next()) {
        QSqlRecord record = query.record();

        time_t timestamp = record.value("time_stamp").toDateTime().toTime_t();
        s.timestamp.append(timestamp);
        s.timestampUnix.append(timestamp);

        if (columns.testFlag(SC_Temperature))
            s.temperature.append(record.value(tempColumn).toDouble());
        if (columns.testFlag(SC_Humidity)) {
            s.humidity.append(record.value(humidityColumn).toDouble());
        }
        if (columns.testFlag(SC_DewPoint)) {
            if (isZxWeather && useIndoorSensors) {
                // ZXweather doesn't store dew point for the indoor sensors.
                // We must calculate it ourselves.
                float temperature = record.value(tempColumn).toDouble();
                float humidity = record.value(humidityColumn).toDouble();
                s.dew_point.append(dewPoint(temperature, humidity));
            } else {
                s.dew_point.append(record.value("dew_point").toDouble());
            }
        }

    }

    return s;
}
