/*****************************************************************************
 *          Copyright: (C) Copyright David Goodwin, 2014
 *            License: GNU General Public License
 *****************************************************************************
 *
 *   This is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This software is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this software; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ****************************************************************************/

#include "mainwindow.h"
#include "ui_mainwindow.h"

#include "temperdatasource.h"
#include "temperhumdatasource.h"
#include "databasedatasource.h"

#include "postgresqldatalogger.h"
#include "sqlitedatalogger.h"

#include "aboutdialog.h"
#include "settingsdialog.h"
#include "constants.h"
#include "settings.h"
#include "chartoptionsdialog.h"
#include "chartwindow.h"
#include "exportdialog.h"

#include <QtDebug>
#include <QMessageBox>
#include <QApplication>

#include <cmath>
#include <algorithm>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    ui->gbDewPoint->hide();
    ui->gbHumidity->hide();

    logData = false;

    // Setup system tray icon
    sysTrayIcon.reset(new QSystemTrayIcon(this));
    sysTrayIcon->setIcon(QIcon(":/icons/app_icon"));
    sysTrayIcon->setToolTip(tr("No data"));
    sysTrayIcon->show();
    connect(sysTrayIcon.data(), SIGNAL(activated(QSystemTrayIcon::ActivationReason)),
            this, SLOT(trayIconActivated(QSystemTrayIcon::ActivationReason)));

    // And the system tray icon menu
    trayIconMenu.reset(new QMenu(this));
    restoreAction.reset(new QAction(tr("&Restore"),trayIconMenu.data()));
    quitAction.reset(new QAction(tr("&Quit"),trayIconMenu.data()));
    trayIconMenu->addAction(restoreAction.data());
    trayIconMenu->addSeparator();
    trayIconMenu->addAction(quitAction.data());
    sysTrayIcon->setContextMenu(trayIconMenu.data());
    connect(restoreAction.data(), SIGNAL(triggered()), this, SLOT(showNormal()));
    connect(quitAction.data(), SIGNAL(triggered()), this, SLOT(quit()));

    // And the main UI
    connect(ui->actionAbout, SIGNAL(triggered()), this, SLOT(about()));
    connect(ui->actionSettings, SIGNAL(triggered()), this, SLOT(settings()));
    connect(ui->actionCreate_Chart, SIGNAL(triggered()), this, SLOT(chart()));
    connect(ui->actionExport, SIGNAL(triggered()), this, SLOT(exportData()));

    currentTs = 0;
    temperatureTotal = 0.0;
    humidityTotal = 0.0;
    dewPointTotal = 0.0;
    samples = 0;

    if (Settings::getInstance().isFirstRun()) {
        settings();
    } else {
        // Then get some data.
        setupLogger();
        setupDataSource();

        // And plot it.
        setupPlot();
    }
}

MainWindow::~MainWindow()
{
    dataSource->closeDevice();
    delete ui;
}

void MainWindow::trayIconActivated(QSystemTrayIcon::ActivationReason reason) {
    switch(reason) {
    case QSystemTrayIcon::Trigger:
    case QSystemTrayIcon::MiddleClick:
    case QSystemTrayIcon::DoubleClick:
        showNormal();
        activateWindow();
    default:
        ;
    }
}

void MainWindow::quit() {
    QCoreApplication::quit();
}

void MainWindow::changeEvent(QEvent *e)
{
    Settings& settings = Settings::getInstance();

    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    case QEvent::WindowStateChange:
        if (windowState().testFlag(Qt::WindowMinimized) &&
                settings.miniseToSysTray()) {

            if (!settings.singleShotMinimiseToSysTray()) {
                QMessageBox::information(this, tr("TempView"),
                                     tr("TempView will minimise to the "
                                     "system tray. To restore it, click on the "
                                     "icon. This behaviour can be changed in the "
                                     "settings dialog."));
                settings.setSingleShotMinimiseToSysTray();
            }

            // We can't call hide from the event handler. So we get the
            // timer to dispatch the hide request for us.
            QTimer::singleShot(0, this, SLOT(hide()));
        }
        break;
    default:
        break;
    }
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    Settings& settings = Settings::getInstance();

    if (sysTrayIcon->isVisible()) {
        if (settings.closeToSysTray()) {
            if (!settings.singleShotCloseToSysTray()) {
                QMessageBox::information(this, tr("TempView"),
                                     tr("TempView will keep running in the "
                                        "system tray. To restore it, click on the "
                                        "icon. To exit, right-click on the system tray "
                                        "icon and choose <b>Exit</b>. This behaviour "
                                        "can be changed from the settings dialog."));
                settings.setSingleShotCloseToSysTray();
            }
            hide();
            event->ignore();
        } else if (settings.logToDatabase() &&
                   (settings.liveDataSourceType() == Settings::DS_TYPE_TEMPER  ||
                    settings.liveDataSourceType() == Settings::DS_TYPE_TEMPERHUM)&&
                   settings.promptOnExitIfLogging()) {

            /*
             * Check that we're actually logging data before warning the user. Its
             * possible we might have failed to open the device which would make
             * any warnings pointless.
             */
            auto ds = qobject_cast<TemperDataSource*>(dataSource.data());
            if (ds != NULL && !ds->isCollecting()) {
                // The data source isn't currently collecting. Nothing to warn about.
                quit();
                return;
            }
            auto ds2 = qobject_cast<TemperHumDataSource*>(dataSource.data());
            if (ds2 != NULL && !ds2->isCollecting()) {
                // The data source isn't currently collecting. Nothing to warn about.
                quit();
                return;
            }


            /*
             * If we're logging to a database from hardware then confirm if the
             * user really wants to exit.
             */
            auto msgBox = new QMessageBox(tr("Really Exit?"),
                                          tr("TempView is currently logging to a database. "
                                          "If you exit then TempView will stop logging. "
                                          "Are you sure you want to exit?"),
                                          QMessageBox::Question,
                                          QMessageBox::Yes,
                                          QMessageBox::Cancel,
                                          QMessageBox::NoButton,
                                          this);
            msgBox->button(QMessageBox::Yes)->setText(tr("E&xit"));
            if (msgBox->exec() == QMessageBox::Yes) {
                quit();
                return;
            }
            event->ignore();
        } else {
            quit();
        }
    }
}

void MainWindow::setupDataSource() {
    ui->actionCreate_Chart->setEnabled(false);

    Settings& settings = Settings::getInstance();

    if (settings.liveDataSourceType() == Settings::DS_TYPE_DATABASE) {
        auto ds = new DatabaseDataSource;
        dataSource.reset(ds);
        ui->actionCreate_Chart->setEnabled(true);
        // TODO: Check database to see if humidity data is available
    } else if (settings.liveDataSourceType() == Settings::DS_TYPE_TEMPERHUM) {
        auto ds = new TemperHumDataSource(settings.temperatureCalibrationOffset(),
                                          settings.humidityCalibrationOffset());
        ds->setInterval(1000);
        dataSource.reset(ds);
        humidityAvailable = true;

        if (Settings::getInstance().logToDatabase())
            ui->actionCreate_Chart->setEnabled(true);
    } else {
        auto ds = new TemperDataSource(settings.temperatureCalibrationOffset());
        ds->setInterval(1000);
        dataSource.reset(ds);
        humidityAvailable = false;

        if (Settings::getInstance().logToDatabase())
            ui->actionCreate_Chart->setEnabled(true);
    }

    connect(dataSource.data(), SIGNAL(newData(float)),
            this, SLOT(dataArrived(float)));
    connect(dataSource.data(), SIGNAL(newTempHumData(float,float)),
            this, SLOT(dataArrived(float, float)));
    connect(dataSource.data(), SIGNAL(messageOutput(QString)),
            this, SLOT(messageReceived(QString)));
    connect(dataSource.data(), SIGNAL(error(QString,QString)),
            this, SLOT(dataSourceError(QString,QString)));

    if (dataSource->openDevice()) {
        dataSource->startCollecting();
    } else {
        qDebug() << "Device open failed. No live data available.";
    }

    ui->actionExport->setEnabled(ui->actionCreate_Chart->isEnabled());
}

void MainWindow::dataSourceError(QString title, QString message) {
    QMessageBox::critical(this, title, message);
}

void MainWindow::setupPlot() {

    auto colours = Settings::getInstance().getChartColours();
    auto graphsEnabled = Settings::getInstance().getMainWindowGraphsEnabled();

    bool singleValueAxis = true;

    ui->plot->clearPlottables();
    ui->plot->clearGraphs();

    ui->plot->axisRect()->setupFullAxesBox();

    ui->plot->yAxis2->setLabel(QString());

    if (graphsEnabled.temperature) {
        // Temperature line
        temperatureGraph = ui->plot->addGraph();
        temperatureGraph->setPen(QPen(colours.temperature));
        temperatureGraph->setName(tr("Temperature"));

        // Temperature dot
        temperaturePointGraph = ui->plot->addGraph();
        temperaturePointGraph->setPen(QPen(colours.temperature));
        temperaturePointGraph->setLineStyle(QCPGraph::lsNone);
        temperaturePointGraph->setScatterStyle(QCPScatterStyle::ssDisc);
        temperaturePointGraph->removeFromLegend();
    }

    if (humidityAvailable) {
        if (graphsEnabled.humidity) {
            // Humidity line
            humidityGraph = ui->plot->addGraph();
            humidityGraph->setPen(QPen(colours.humidity));
            humidityGraph->setName(tr("Humidity"));

            // Humidity dot
            humidityPointGraph = ui->plot->addGraph();
            humidityPointGraph->setPen(QPen(colours.humidity));
            humidityPointGraph->setLineStyle(QCPGraph::lsNone);
            humidityPointGraph->setScatterStyle(QCPScatterStyle::ssDisc);
            humidityPointGraph->removeFromLegend();
        }

        if (graphsEnabled.dewPoint) {
            // Dew point line
            dewPointGraph = ui->plot->addGraph();
            dewPointGraph->setPen(QPen(colours.dewPoint));
            dewPointGraph->setName(tr("Dew Point"));

            // Dew point dot
            dewPointPointGraph = ui->plot->addGraph();
            dewPointPointGraph->setPen(QPen(colours.dewPoint));
            dewPointPointGraph->setLineStyle(QCPGraph::lsNone);
            dewPointPointGraph->setScatterStyle(QCPScatterStyle::ssDisc);
            dewPointPointGraph->removeFromLegend();
        }
    }

    if ((temperatureGraph || dewPointGraph) && humidityGraph) {
        // Configure value axes given we've now got more than one
        if (temperatureGraph) {
            temperatureGraph->setValueAxis(ui->plot->yAxis);
            temperaturePointGraph->setValueAxis(ui->plot->yAxis);
        }

        humidityGraph->setValueAxis(ui->plot->yAxis2);
        humidityPointGraph->setValueAxis(ui->plot->yAxis2);

        if (dewPointGraph) {
            dewPointGraph->setValueAxis(ui->plot->yAxis);
            dewPointPointGraph->setValueAxis(ui->plot->yAxis);
        }

        ui->plot->yAxis->setLabel(tr("Temperature"));
        ui->plot->yAxis2->setLabel(tr("Humidity"));
        ui->plot->yAxis2->setTickLabels(true);
        ui->plot->yAxis2->setAutoTickLabels(true);
        singleValueAxis = false;
    } else if (humidityGraph) {
        // Humidity is the only graph enabled

        humidityGraph->setValueAxis(ui->plot->yAxis);
        humidityPointGraph->setValueAxis(ui->plot->yAxis);

        ui->plot->yAxis->setLabel(tr("Humidity"));
        ui->plot->yAxis2->setLabel(QString());
    } else {
        // One of the temperature graphs is the only one enabled

        ui->plot->yAxis->setLabel(tr("Temperature"));

        if (temperatureGraph) {
            temperatureGraph->setValueAxis(ui->plot->yAxis);
            temperaturePointGraph->setValueAxis(ui->plot->yAxis);
        } else if (dewPointGraph) {
            dewPointGraph->setValueAxis(ui->plot->yAxis);
            dewPointPointGraph->setValueAxis(ui->plot->yAxis);
        }
    }

    int numberGraphsEnabled = 0;
    if (temperatureGraph) numberGraphsEnabled++;
    if (humidityGraph) numberGraphsEnabled++;
    if (dewPointGraph) numberGraphsEnabled++;

    if (ui->plot->legend) {
        ui->plot->legend->setVisible(
                    numberGraphsEnabled > 1 &&
                    Settings::getInstance().mainWindowGraphLegendEnabled());
        ui->plot->axisRect()->insetLayout()->setInsetAlignment(0, Qt::AlignTop
                                                                   | Qt::AlignLeft);
    }

    // Configure axes
    ui->plot->xAxis->setTickLabelType(QCPAxis::ltDateTime);
    ui->plot->xAxis->setDateTimeFormat(tr("hh:mm:ss"));


    // Link axis ranges
    connect(ui->plot->xAxis, SIGNAL(rangeChanged(QCPRange)),
            ui->plot->xAxis2, SLOT(setRange(QCPRange)));

    if (singleValueAxis) {
        connect(ui->plot->yAxis, SIGNAL(rangeChanged(QCPRange)),
                ui->plot->yAxis2, SLOT(setRange(QCPRange)));
    }
}

void MainWindow::dataArrived(float temperature) {

    if (humidityAvailable) {
        // Seems we've changed data sources! Most likely we've just connected
        // to a database that isn't broadcasting humidity. Better reconfigure
        // the chart...
        humidityAvailable = false;
        setupPlot();
    }

    ui->lblTemperature->setText(tr("%1°C").arg(QString::number(temperature, 'f', 1)));

    // This data source doesn't provide humidity. Turn it and anything derived
    // from it off.
    ui->gbHumidity->hide();
    ui->gbDewPoint->hide();

    if (Settings::getInstance().mainWindowGraphAveraged()) {
        averageData(temperature, 0, 0);
    } else {
        updatePlot(temperature);
    }

    if (logData)
        dataLogger->logData(temperature);

    QString formatString = tr("Temperature: %1°C");
    QString iconText = formatString
            .arg(QString::number(temperature, 'f', 1));
    sysTrayIcon->setToolTip(iconText);
}

void MainWindow::dataArrived(float temperature, float humidity) {

    if (!humidityAvailable) {
        // Seems we've changed data sources! Most likely we've just connected
        // to a database that is broadcasting humidity. Better reconfigure
        // the chart...
        humidityAvailable = true;
        setupPlot();
    }

    auto dp = dewPoint(temperature, humidity);

    ui->lblTemperature->setText(tr("%1°C").arg(QString::number(temperature, 'f', 1)));

    ui->lblHumidity->setText(tr("%1%").arg(QString::number(humidity, 'f', 1)));
    ui->gbHumidity->show();

    ui->lblDewPoint->setText(tr("%1°C").arg(QString::number(dp, 'f', 1)));
    ui->gbDewPoint->show();

    if (Settings::getInstance().mainWindowGraphAveraged()) {
        averageData(temperature, humidity, dewPoint(temperature, humidity));
    } else {
        updatePlot(temperature, humidity, dewPoint(temperature, humidity));
    }

    if (logData)
        dataLogger->logData(temperature, humidity, dp);

    QString formatString = tr("Temperature: %1°C\nHumidity: %2%\nDew Point: %3°C");
    QString iconText = formatString
            .arg(QString::number(temperature, 'f', 1))
            .arg(QString::number(humidity, 'f', 1))
            .arg(QString::number(dp, 'f', 1));
    sysTrayIcon->setToolTip(iconText);
}

void MainWindow::updatePlot(float temperature) {
    updatePlot(temperature, 0, 0);
}

void MainWindow::updatePlot(float temperature, float humidity, float dewPoint) {
    // Update graph
    double key = QDateTime::currentDateTime().toMSecsSinceEpoch()/1000.0;

    bool averaged = Settings::getInstance().mainWindowGraphAveraged();

    qint64 xRange = averaged ? Constants::AVERAGED_CHART_RANGE : Constants::CHART_RANGE;

    // The temperature line
    if (temperatureGraph) {
        temperatureGraph->addData(key, temperature);

        // clip graph (throw away data that falls outside the range)
        temperatureGraph->removeDataBefore(key - xRange);
    }

    // The temperature dot
    if (temperaturePointGraph) {
        temperaturePointGraph->clearData();
        temperaturePointGraph->addData(key, temperature);
    }

    // The humidity line
    if (humidityGraph) {
        humidityGraph->addData(key, humidity);

        // clip graph (throw away data that falls outside the range)
        humidityGraph->removeDataBefore(key - xRange);
    }

    // The humidity dot
    if (humidityPointGraph) {
        humidityPointGraph->clearData();
        humidityPointGraph->addData(key, humidity);
    }

    // The dew point line
    if (dewPointGraph) {
        dewPointGraph->addData(key, dewPoint);

        // clip graph (throw away data that falls outside the range)
        dewPointGraph->removeDataBefore(key - xRange);
    }

    // The dew point dot
    if (dewPointPointGraph) {
        dewPointPointGraph->clearData();
        dewPointPointGraph->addData(key, dewPoint);
    }



    // Make axes scroll with the data with a constant range
    ui->plot->xAxis->setRange(key + (averaged ?
                                         Constants::AVERAGED_RIGHT_PADDING :
                                         Constants::RIGHT_PADDING),
                              xRange,
                              Qt::AlignRight);

    QCPRange yRange;

    if (temperatureGraph || dewPointGraph) {
        // Temperature is on the Left value axis, humidity on the right
        auto oldTempRange = ui->plot->yAxis->range();

        if (temperatureGraph && dewPointGraph) {
            temperatureGraph->rescaleValueAxis();
            dewPointGraph->rescaleValueAxis(true);
        } else if (temperatureGraph) {
            temperatureGraph->rescaleValueAxis();
        } else if (dewPointGraph) {
            dewPointGraph->rescaleValueAxis();
        }
        yRange = ui->plot->yAxis->range();

        /*
         * Apply some padding to the value axis so the line is always visible.
         * Only do this if the range was changed during rescale though (as the
         * rescale will have removed the padding we put on last time)
         */

        if (yRange.lower != oldTempRange.lower || yRange.upper != oldTempRange.upper) {
            ui->plot->yAxis->setRange(yRange.lower - Constants::VALUE_AXIS_PADDING,
                                      yRange.upper + Constants::VALUE_AXIS_PADDING);
        }

        if (humidityGraph) {
            auto oldHumRange = ui->plot->yAxis2->range();
            humidityGraph->rescaleValueAxis();

            auto yHumRange = ui->plot->yAxis2->range();
            if (yHumRange.lower != oldHumRange.lower || yHumRange.upper != oldHumRange.upper) {
                ui->plot->yAxis2->setRange(yHumRange.lower - Constants::VALUE_AXIS_PADDING,
                                          yHumRange.upper + Constants::VALUE_AXIS_PADDING);
            }
        }
    } else if (humidityGraph) {
        // Humidity is on the left value axis.
        auto oldHumRange = ui->plot->yAxis->range();

        humidityGraph->rescaleValueAxis();

        auto yHumRange = ui->plot->yAxis->range();
        if (yHumRange.lower != oldHumRange.lower || yHumRange.upper != oldHumRange.upper) {
            ui->plot->yAxis->setRange(yHumRange.lower - Constants::VALUE_AXIS_PADDING,
                                      yHumRange.upper + Constants::VALUE_AXIS_PADDING);
        }
    }

    ui->plot->replot();
}

void MainWindow::messageReceived(QString message) {
    qDebug() << message;
}

void MainWindow::about() {
    AboutDialog ad;
    ad.exec();
}

void MainWindow::settings() {

    auto oldColours = Settings::getInstance().getChartColours();
    auto oldGraphs = Settings::getInstance().getMainWindowGraphsEnabled();
    auto oldDs = Settings::getInstance().liveDataSourceType();

    SettingsDialog sd;
    int result = sd.exec();
    if (result == QDialog::Rejected) {
        return; // User canceled
    }

    // Only reset the plot if the colour or other graph settings has been changed
    // (We don't want to reset it every time like the DS as it clears
    // the chart)
    auto colours = Settings::getInstance().getChartColours();
    auto graphs = Settings::getInstance().getMainWindowGraphsEnabled();
    auto ds = Settings::getInstance().liveDataSourceType();

    auto colourChanged = (colours.temperature == oldColours.temperature) &&
            (colours.humidity == oldColours.humidity) &&
            (colours.dewPoint == oldColours.dewPoint);

    auto graphsChanged = (graphs.temperature == oldGraphs.temperature) &&
            (graphs.humidity == oldGraphs.humidity) &&
            (graphs.dewPoint == oldGraphs.dewPoint);

    // Reconnect to the datasource in case it was changed.
    setupDataSource();
    setupLogger();

    if (colourChanged || graphsChanged || oldDs != ds) {
        setupPlot();
    }
}

void MainWindow::setupLogger() {

    Settings& settings = Settings::getInstance();
    if ((settings.liveDataSourceType() == Settings::DS_TYPE_TEMPER
         || settings.liveDataSourceType() == Settings::DS_TYPE_TEMPERHUM)
            && settings.logToDatabase()) {

        if (settings.databaseDriver() == "QPSQL") {
            dataLogger.reset(new PostgreSQLDataLogger());
        } else {
            dataLogger.reset(new SQLiteDataLogger());
        }

        connect(dataLogger.data(), SIGNAL(error(QString,QString)),
                this, SLOT(dataSourceError(QString,QString)));

        dataLogger->connectLogger();

        logData = true;
    } else {

        logData = false;
    }
}

SampleSet MainWindow::GetSamples(QDateTime startTime, QDateTime endTime, SampleColumns columns)
{
    SampleSet samples;

    /*
     * We need to get samples from somewhere. There are two completely different
     * and unrelated things that can provide them:
     *   -> The Database Data Source
     *   -> The Data Logger
     */

    AbstractDataSource *ads = dataSource.data();
    DatabaseDataSource *dds = dynamic_cast<DatabaseDataSource*>(ads);

    Q_ASSERT_X(dds != NULL || dataLogger.data() != NULL,
               "chart()", "No sample source");

    if (dds != NULL) {
        /*
         * Were receiving data from a remote PostgreSQL database (no local data
         * acquisition or logging). We need to fetch the samples from there.
         */
        samples = dds->fetchSamples(columns, startTime, endTime);
    } else {
        /*
         * We're acquiring data locally and logging it. We need to get data from
         * where ever we're logging it to.
         */
        samples = dataLogger->fetchSamples(columns, startTime, endTime);
    }

    return samples;
}

void MainWindow::chart() {
    ChartOptionsDialog co(this);
    if (co.exec() == QDialog::Accepted) {
        // Display chart window.

        SampleSet samples = GetSamples(co.getStartTime(), co.getEndTime(),
                                       co.getColumns());

        ChartWindow *cw = new ChartWindow(samples, co.getColumns());
        cw->setAttribute(Qt::WA_DeleteOnClose);
        cw->show();
    }
}

void MainWindow::exportData() {
    ExportDialog ed(this);
    if (ed.exec() == QDialog::Accepted) {

        SampleSet samples = GetSamples(ed.getStartTime(), ed.getEndTime(), ed.getColumns());

        ExportDialog::performExport(samples,
                                    this,
                                    ed.getColumns(),
                                    ed.getTargetFilename(),
                                    ed.getDelimiter(),
                                    ed.includeHeader());
    }
}

float MainWindow::dewPoint(float temperature, float humidity) {
    float a = 17.271;
    float b = 237.7; // in °C
    float gamma, result;

    gamma = ((a * temperature) / (b + temperature)) + log(humidity / 100.0);
    result = (b * gamma) / (a - gamma);
    return result;
}


qint64 MainWindow::nextTs() {
    return ceil(QDateTime::currentDateTime().toMSecsSinceEpoch() / 300000.0) * 300000;
}

void MainWindow::averageData(float temperature, float humidity, float dewPoint) {

    auto ts = nextTs();

    if (ts > currentTs) {
        // Time to average it.
        if (samples == 0) {
            // Put the first point in the plot
            updatePlot(temperature, humidity, dewPoint);
        } else {
            float averageTemperature = temperatureTotal / samples;
            float averageHumidity = humidityTotal / samples;
            float averageDewPoint = dewPointTotal / samples;

            // Then store it
            updatePlot(averageTemperature, averageHumidity, averageDewPoint);
        }

        // Then reset everything for the next time period.
        temperatureTotal = temperature;
        humidityTotal = humidity;
        dewPointTotal = dewPoint;
        samples = 1;
        currentTs = ts;
        qDebug() << "New TS:" << QDateTime::fromMSecsSinceEpoch(ts);
    } else {
        temperatureTotal += temperature;
        humidityTotal += humidity;
        dewPointTotal += dewPoint;
        samples++;
    }
}
