/*****************************************************************************
 *          Copyright: (C) Copyright David Goodwin, 2014
 *            License: GNU General Public License
 *****************************************************************************
 *
 *   This is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This software is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this software; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ****************************************************************************/

#include "sqlitedatalogger.h"
#include "settings.h"

#include <QtDebug>
#include <QFile>
#include <QDir>
#include <QSqlDatabase>
#include <QSqlQuery>
#include <QSqlRecord>
#include <QSqlError>


#define SQLITE_DATALOGGER_DB "sqlite-datalogger"
#define SQLiteDb QSqlDatabase::database(SQLITE_DATALOGGER_DB)

SQLiteDataLogger::SQLiteDataLogger(QObject *parent) :
    AbstractDataLogger(parent)
{
    isConnected = false;
}

SQLiteDataLogger::~SQLiteDataLogger() {
    disconnectLogger();
}

bool SQLiteDataLogger::connectLogger() {
    auto db = QSqlDatabase::addDatabase("QSQLITE", SQLITE_DATALOGGER_DB);

    QString filename = Settings::getInstance().databaseFileName();

    // Ensure the path to the database file exists.
//    if (!QDir(QDir(filename).path()).exists())
//        QDir().mkpath(QDir(filename).path());

    qDebug() << "Database file:" << filename;

    db.setDatabaseName(filename);

    if (!db.open()) {
        emit error(tr("Datbase Error"),
                   QString(tr("Failed to open database file. %1")).arg(
                       db.lastError().text()));
        return false;
    }

    QSqlQuery query("select * from sqlite_master "
                    "where name='db_info' and type='table'",
                    SQLiteDb);
    if (!query.isActive()) {
        // TODO: Just trash the database and re-create it in this case.
        qWarning() << "Failed to determine cache version. Cache functionality "
                      "will be disabled.";
        emit error(tr("Database Error"), tr("Failed to determine db version: Check for db_info failed."));
        return false;
    }

    if (!query.next()) {
        createDatabaseStructure(); // Its a blank database
    }

    int dbVersion = getDatabaseVersion();
    if (dbVersion < 1) {
        emit error(tr("Database Error"), tr("Failed to determine db version."));
        return false;
    } else if (dbVersion == 1) {
        // We need to upgrade the database to schema level 2. Level 2 is
        // backwards compatible with level 1 (TempView 1.x) so we wont bother
        // asking permission.
        upgradeDatabaseStructure();
    }

    if (!AbstractDataLogger::connectLogger())
        return false;

    isConnected = true;
    return true;
}

int SQLiteDataLogger::getDatabaseVersion() {
    QSqlQuery query(SQLiteDb);
    query.exec("select v "
               "from DB_INFO "
               "where k = 'DB_VERSION'");
    if (!query.isActive()) {
        return -1;
    } else {
        query.first();
        return query.value(0).toString().toInt();
    }

    return 1;
}


void SQLiteDataLogger::disconnectLogger() {
    isConnected = false;
    auto db = SQLiteDb;

    db.close();
    QSqlDatabase::removeDatabase(SQLITE_DATALOGGER_DB);
}

void SQLiteDataLogger::insertSample(qint64 timestamp, float temperature, float humidity, float dewPoint) {
    if (isConnected) {
        if (humidity > 0) {
            QSqlQuery qry(SQLiteDb);
            qry.prepare("insert into sample(time_stamp, temperature, humidity, "
                        "dew_point) values(:ts, :temp, :hum, :dp)");
            qry.bindValue(":ts", QDateTime::fromMSecsSinceEpoch(timestamp).toTime_t());
            qry.bindValue(":temp", temperature);
            qry.bindValue(":hum", humidity);
            qry.bindValue(":dp", dewPoint);
            if (!qry.exec()) {
                qWarning() << "Failed to log data";
                qDebug() << qry.lastError().databaseText();
                qDebug() << qry.lastError().driverText();
            }
        } else {
            // No humidity or dew point
            QSqlQuery qry(SQLiteDb);
            qry.prepare("insert into sample(time_stamp, temperature) "
                        "values(:ts, :temp)");
            qry.bindValue(":ts", QDateTime::fromMSecsSinceEpoch(timestamp).toTime_t());
            qry.bindValue(":temp", temperature);
            if (!qry.exec()) {
                qWarning() << "Failed to log data";
                qDebug() << qry.lastError().databaseText();
                qDebug() << qry.lastError().driverText();
            }
        }
    }
}

SampleSet SQLiteDataLogger::fetchSamples(SampleColumns columns, QDateTime start, QDateTime end) {
    SampleSet s;

    QString columnList = "";
    if (columns.testFlag(SC_Temperature)) columnList += ", temperature ";
    if (columns.testFlag(SC_Humidity))    columnList += ", humidity ";
    if (columns.testFlag(SC_DewPoint))   columnList += ", dew_point ";

    QString queryText =
            "select time_stamp " + columnList +
            "from sample "
            "where time_stamp >= :startTs "
            "  and time_stamp <= :endTs "
            "order by time_stamp";

    QSqlQuery query(SQLiteDb);
    query.prepare(queryText);
    query.bindValue(":startTs", start.toTime_t());
    query.bindValue(":endTs", end.toTime_t());

    bool result = query.exec();

    if (!result || !query.isActive()) {
        emit error(tr("Database Error"), query.lastError().driverText());
        return s;
    }

    while (query.next()) {
        QSqlRecord record = query.record();

        time_t timestamp = record.value("time_stamp").toUInt();
        s.timestamp.append(timestamp);
        s.timestampUnix.append(timestamp);

        if (columns.testFlag(SC_Temperature))
            s.temperature.append(record.value("temperature").toDouble());
        if (columns.testFlag(SC_Humidity))
            s.humidity.append(record.value("humidity").toDouble());
        if (columns.testFlag(SC_DewPoint))
            s.dew_point.append(record.value("dew_point").toDouble());
    }

    return s;
}

void SQLiteDataLogger::createDatabaseStructure() {
    runDatabaseScript(":/db/init.sql",
                      tr("Failed to create structure. Error was: %1"));
}

void SQLiteDataLogger::upgradeDatabaseStructure() {
    runDatabaseScript(":/db/upgrade.sql",
                      tr("Failed to upgrade structure. Error was: %1"));
}

bool SQLiteDataLogger::runDatabaseScript(QString scriptFileName,
                                         QString errorTemplate) {
    qDebug() << "Run script:" << scriptFileName;
    QFile scriptFile(scriptFileName);
    scriptFile.open(QIODevice::ReadOnly);

    QString script;

    while(!scriptFile.atEnd()) {
        QString line = scriptFile.readLine();
        if (!line.startsWith("--"))
            script += line;
    }

    // This is really nasty.
    QStringList statements = script.split(";");

    QSqlQuery query(SQLiteDb);

    foreach(QString statement, statements) {
        statement = statement.trimmed();

        if (statement == "")
            continue;

        qDebug() << statement;

        // The split above strips off the semicolons.
        query.exec(statement);

        if (query.lastError().isValid()) {
            emit error(tr("Database Error"),
                       errorTemplate.arg(
                         query.lastError().driverText()));
            qWarning() << query.lastError().text();
            return false;
        }
    }

    return true;
}
